--------------------------------------------------------------------------------
-- Copyright(C) 2005 by Xilinx, Inc. All rights reserved.
-- This text/file contains proprietary, confidential
-- information of Xilinx, Inc., is distributed under license
-- from Xilinx, Inc., and may be used, copied and/or
-- disclosed only pursuant to the terms of a valid license
-- agreement with Xilinx, Inc. Xilinx hereby grants you
-- a license to use this text/file solely for design, simulation,
-- implementation and creation of design files limited
-- to Xilinx devices or technologies. Use with non-Xilinx
-- devices or technologies is expressly prohibited and
-- immediately terminates your license unless covered by
-- a separate agreement.
--
-- Xilinx is providing this design, code, or information
-- "as is" solely for use in developing programs and
-- solutions for Xilinx devices. By providing this design,
-- code, or information as one possible implementation of
-- this feature, application or standard, Xilinx is making no
-- representation that this implementation is free from any
-- claims of infringement. You are responsible for
-- obtaining any rights you may require for your implementation.
-- Xilinx expressly disclaims any warranty whatsoever with
-- respect to the adequacy of the implementation, including
-- but not limited to any warranties or representations that this
-- implementation is free from claims of infringement, implied
-- warranties of merchantability or fitness for a particular
-- purpose.
--
-- Xilinx products are not intended for use in life support
-- appliances, devices, or systems. Use in such applications are
-- expressly prohibited.
--
-- This copyright and support notice must be retained as part
-- of this text at all times. (c) Copyright 2005 Xilinx, Inc.
-- All rights reserved.
--------------------------------------------------------------------------------
--
-- Filename - filter_core.vhd
-- Author - P. Szanto
-- Creation - 27 June 2006
--
-- Description -
--
-- The main part of the 2D Rank filter, the filter core.
-- It can accept NI new inputs in a single clock cycle.
-- The output is NI addresses for the NI delay lines and a multiplexer control signal to select the appropriate delay line.
-- Latency is configuration dependant.
-- Sub-modules are comparators and bit-sum modules (cnt1_xxx)
--
--
-- $RCSfile: filter_core.vhd,v $ $Revision: 1.9 $ $Date: 2006/06/30 19:01:22 $
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

use work.rank2d_utils.ALL;
use work.rank2d_latency.ALL;


entity filter_core is
generic(DW        : integer := 10;        -- data width
        TAPr      : integer := 9;         -- real tap number
        TAPb      : integer := 6;         -- log2TAP, upper integer
        WH_real   : integer := 1;         -- filter window vertical resolution (real)
        NI        : integer := 1;         -- new samples/clk
        ADDER_INS : integer := 4;
        FAMILY    : string  := "Virtex4";
        USE_INST_COMP : integer := 1);
port   (clk       : in  std_logic;
        rst       : in  std_logic;
        ce        : in  std_logic;
        pix_en    : in  std_logic;
        m_ni_en   : in  std_logic;
        m_cell_en : in  std_logic;
        rank      : in  std_logic_vector(TAPb-1 downto 0);
        nd_port   : in  std_logic_vector(NI*DW-1 downto 0);
        out_addr  : out std_logic_vector(NI*TAPb-1 downto 0);
        out_mux   : out std_logic_vector(LOG2_CEIL(NI)-1 downto 0)
      );
--attribute syn_maxfan : integer;
--attribute syn_maxfan of pix_en : signal is 40;
--attribute syn_direct_enable : boolean;
--attribute syn_direct_enable of m_ni_en, m_cell_en : signal is true;
end filter_core;

architecture rtl of filter_core is

-- attributes
attribute syn_useenables : boolean;
attribute syn_direct_enable : boolean;
attribute syn_pipeline : boolean;
attribute register_balancing: string;
attribute syn_srlstyle : string;
attribute shreg_extract : string;
attribute syn_maxfan : integer;

-- constants
constant CONST_0        : std_logic_vector(127 downto 0) := x"00000000000000000000000000000000";
constant CONST_1        : std_logic_vector(127 downto 0) := x"FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF";
constant WH             : integer := TAPr/WH_real;
constant WH_virt        : integer := DIV_CEIL(WH_real, NI)*NI;
constant TAPv           : integer := WH_virt*WH;
constant mask_num       : integer := DIV_CEIL(WH_real, NI);
constant TAP            : integer := (TAPr/WH_real)*WH_virt;
constant NZ             : integer := WH_virt - WH_real;
constant sub_ops_num    : integer := (TAP-NI)*NI + NI*(NI-1)/2;
constant cnt1_latency   : integer := GET_1CNT_STAGES(TAP, ADDER_INS) + 2;
constant out_latency    : integer := cnt1_latency + 3;
constant RANK_PER_STAGE : integer := 8;
constant ranking_regs   : integer := DIV_CEIL(TAPr, rank_per_stage);
constant cnt1_valid_lat : integer := (WH_virt/NI)*WH + 3 + GET_1CNT_STAGES(TAP, ADDER_INS);
constant RR_PER_NI      : integer := DIV_CEIL(TAP, (NI*RANK_PER_STAGE));
constant mask_ni_rst    : std_logic_vector(WH_virt downto 0) := CONST_1(WH_virt-NZ-1 downto 0) & CONST_0(NZ downto 0);

constant MUX_OR_INS     : integer := 8;
constant MUX0_REGS      : integer := DIV_CEIL(TAP, MUX_OR_INS);

impure function gen_MASK_CELL_INIT (TAP, WH_virt, WH_real: integer) return std_logic_vector is
variable MASK_CELL_INIT : std_logic_vector(TAP-1 downto 0);
begin                                                        
   for I in 0 to TAP-1 loop
      if ((I mod WH_virt) < (WH_virt-WH_real)) then
         MASK_CELL_INIT(I) := '0';
      else
         MASK_CELL_INIT(I) := '1';
      end if;
   end loop; 
   return MASK_CELL_INIT;
end function;                                                

constant MASK_CELL_INIT : std_logic_vector(TAP-1 downto 0) := gen_MASK_CELL_INIT(TAP, WH_virt, WH_real);

type array_NI_by_DW      is array (NI-1 downto 0) of STD_LOGIC_VECTOR (DW-1 downto 0);  
type array_TAP_by_DW     is array (TAP-1 downto 0) of STD_LOGIC_VECTOR (DW-1 downto 0);  
type array_NI_by_NI      is array (NI-1 downto 0) of STD_LOGIC_VECTOR (NI-1 downto 0); 
type array_TAP_by_NI     is array (TAP-1 downto 0) of STD_LOGIC_VECTOR (NI-1 downto 0); 
type array_TAP_by_TAP    is array (TAP-1 downto 0) of STD_LOGIC_VECTOR (TAP-1 downto 0); 
type array_TAP_by_TAPb   is array (TAP-1 downto 0) of STD_LOGIC_VECTOR (TAPb-1 downto 0);
type array_TAPr_by_TAPb  is array (TAPr-1 downto 0) of STD_LOGIC_VECTOR (TAPb-1 downto 0);
type array_NI_DWp        is array (NI-1 downto 0) of std_logic_vector(DW downto 0);
type array_NI_NI_DWp     is array (NI-1 downto 0) of array_NI_DWp;
type array_TAP_NI_DWp    is array (TAP-1 downto 0) of array_NI_DWp;
type array_NI_by_INTEGER is array (NI-1 downto 0) of integer;
type array_rank0_t0      is array (RR_PER_NI-1 downto 0) of STD_LOGIC_VECTOR(TAPb-1 downto 0);
type array_rank0_t1      is array (NI-1 downto 0) of array_rank0_t0;
type array_rank1         is array (NI-1 downto 0) of STD_LOGIC_VECTOR(TAPb-1 downto 0);
type array_SUB_OPS       is array (sub_ops_num-1 downto 0) of std_logic_vector(DW downto 0);
type array_mux_sel       is array (MUX0_REGS-1 downto 0) of STD_LOGIC_VECTOR(LOG2_CEIL(NI)-1 downto 0);

-- LUT RAMs for bit/cell masking
signal mask_ni_shr   : std_logic_vector(WH_virt-1 downto 0) := mask_ni_rst(WH_virt downto 1);
signal mask_ni       : std_logic_vector(NI-1 downto 0);
signal mask_cell_shr : std_logic_vector(TAP-1 downto 0) := MASK_CELL_INIT;
signal mask_cell     : std_logic_vector(TAP-1 downto 0);
attribute syn_useenables of mask_ni_shr : signal is false;
attribute syn_useenables of mask_cell_shr : signal is false;


signal new_data      : array_NI_by_DW;
--attribute syn_maxfan of new_data : signal is 5;-- ok for 551, no for 772?

signal data_reg      : array_TAP_by_DW;
signal cmpr_out_new  : array_NI_by_NI;
signal cmpr_out_old  : array_TAP_by_NI;
signal c_shreg       : array_TAP_by_TAP;
attribute syn_srlstyle of c_shreg : signal is "select_srl";

--- implement subtractor

signal cmpr_new_sub : array_NI_NI_DWp;
signal cmpr_old_sub : array_TAP_NI_DWp;
signal sub_ops_a : array_SUB_OPS;
signal sub_ops_b : array_SUB_OPS;
signal sub_res   : array_SUB_OPS;
attribute syn_pipeline of sub_ops_a : signal is true;
attribute syn_pipeline of sub_ops_b : signal is true;
attribute register_balancing of sub_ops_a: signal is "yes";
attribute register_balancing of sub_ops_b: signal is "yes";


signal masked_full   : array_TAP_by_TAP;
signal masked_in     : array_TAP_by_NI;
signal masked_out    : array_TAP_by_NI;
signal one_cntr      : array_TAP_by_TAPb;
signal rank_cmpr     : std_logic_vector(TAP-1 downto 0);

signal ranked_cell0 : array_rank0_t1;
attribute syn_srlstyle of ranked_cell0 : signal is "registers";
attribute shreg_extract of ranked_cell0 : signal is "no";

signal mux_sel0 : array_mux_sel;
signal mux_sel1 : STD_LOGIC_VECTOR(LOG2_CEIL(NI)-1 downto 0);

signal ranked_cell1   : array_rank1;

begin

-- mask memories, address generator
-- at startup system is resetted to a "completely valid filter window", therefore
-- mask_cell (eg. 5*5/2) is "111110", while mask_ni is "10"
-- mask_cell is synchronized with SH (new 1cnt uses SH data)
-- mask_ni is one cycle earlier, because it uses the older (thrown out) SH and CMPR result
-- mask for the input/thrown away comparison results
-- typically 111...111, except the last sampling in a given column
process(clk)
begin
if (clk'event and clk='1') then
   if (rst='1') then
      mask_ni_shr <= mask_ni_rst(WH_virt downto 1);
   elsif (pix_en='1' AND m_ni_en='1') then
      mask_ni_shr <= mask_ni_shr(WH_virt-NI-1 downto 0) & mask_ni_shr(WH_virt-1 downto WH_virt-NI);
   end if;
end if;
end process;
GEN_MASK_NI_1:
if (NZ=0) generate
   mask_ni <= CONST_1(NI-1 downto 0);
end generate;
GEN_MASK_NI_SHR:
if (NZ/=0) generate
   mask_ni <= mask_ni_shr(NI-1 downto 0);
end generate;


process(clk)
begin
   if (clk'event and clk='1') then
      if (rst='1') then
         mask_cell_shr <= MASK_CELL_INIT;
      elsif (pix_en='1' AND m_cell_en='1') then 
         mask_cell_shr <= mask_cell_shr(TAP-NI-1 downto 0) & mask_cell_shr(TAP-1 downto TAP-NI);
      end if;
   end if;
end process;
GEN_MASK_CELL_1:
if (NZ=0) generate
   mask_cell <= CONST_1(TAP-1 downto 0);
end generate;
GEN_MASK_CELL_SHR:
if (NZ/=0) generate
   mask_cell <= mask_cell_shr;
end generate;
-- end mask generation


-- shift register array for new data
GEN_NEW_DATA:
   for I in NI-1 downto 0 generate
   begin
      process(clk)
      begin
      if (clk'event and clk='1') then
         if (pix_en='1') then
            new_data(I) <= nd_port((I+1)*DW-1 downto I*DW);
         end if;
      end if;
      end process;
   end generate;

-- data registers for the samples within the filter kernel
GEN_DATA_REG0:
   for I in 0 to NI-1 generate
   begin
      process(clk)
      begin
      if (clk'event and clk='1') then      
         if (pix_en='1') then
            data_reg(I) <= new_data(I);
         end if;   
      end if;
      end process;
   end generate;
GEN_DATA_REG1:
   for I in NI to TAP-1 generate
   begin
      process(clk)
      begin
      if (clk'event and clk='1') then      
         if (pix_en='1') then
            data_reg(I) <= data_reg(I-NI);
         end if;
      end if;
      end process;
   end generate;

-- generate SUB array
-- first elements are for old-new subs, others for new-new
-- comparison: new < old
GEN_SUB_OPS_OLD:
for INS in 0 to NI-1 generate
   L1:
   for IT in 0 to TAP-NI-1 generate
      sub_ops_a(INS*(TAP-NI)+IT) <= '0' & new_data(INS);
      sub_ops_b(INS*(TAP-NI)+IT) <= '0' & data_reg(IT);
   end generate;
end generate;
GEN_SUB_OPS_NEW:
for IN0 in 0 to NI-2 generate
   L1: 
   for IN1 in IN0+1 to NI-1 generate
      sub_ops_a( ((TAP-NI)*NI)+ (NI*IN0-(IN0*IN0+IN0)/2) + IN1-IN0-1) <= '0' & new_data(IN0);
      sub_ops_b( ((TAP-NI)*NI)+ (NI*IN0-(IN0*IN0+IN0)/2) + IN1-IN0-1) <= '0' & new_data(IN1);
   end generate;
end generate;

GEN_SUBS:
for ISUB in 0 to ( (TAP-NI)*NI + NI*(NI-1)/2 - 1 ) generate
   COMP_HDL:
   if (USE_INST_COMP=0) generate
      comp: entity work.comp_module(rtl)
            generic map(
                  DW => DW,
                  NUM => ISUB
            )
            port map(
                  clk    => clk,
                  en     => pix_en,
                  op_a   => sub_ops_a(ISUB),
                  op_b   => sub_ops_b(ISUB),
                  res    => sub_res(ISUB)
            );
   end generate;
   COMP_PRIM:
   if (USE_INST_COMP=1) generate
      COMP_V2:
      if (FAMILY="virtex2" or FAMILY="virtex2p") generate
         comp: entity work.sub_inst_v2(struct)
               generic map(
                     DW => DW,
                     NUM => ISUB
               )
               port map(
                     clk    => clk,
                     en     => pix_en,
                     op_a   => sub_ops_a(ISUB),
                     op_b   => sub_ops_b(ISUB),
                     res    => sub_res(ISUB)
               );
      end generate;
      COMP_V4:
      if (FAMILY="virtex4") generate
         comp: entity work.sub_inst_v4(struct)
               generic map(
                     DW => DW,
                     NUM => ISUB
               )
               port map(
                     clk    => clk,
                     en     => pix_en,
                     op_a   => sub_ops_a(ISUB),
                     op_b   => sub_ops_b(ISUB),
                     res    => sub_res(ISUB)
               );         
      end generate;
      COMP_V5:
      if (FAMILY="virtex5") generate
         comp: entity work.comp_module(rtl)
               generic map(
                     DW => DW,
                     NUM => ISUB
               )
               port map(
                     clk    => clk,
                     en     => pix_en,
                     op_a   => sub_ops_a(ISUB),
                     op_b   => sub_ops_b(ISUB),
                     res    => sub_res(ISUB)
               );   
      end generate;
      COMP_OTHERS:   
      if (FAMILY/="virtex5" and FAMILY/="virtex4" and FAMILY/="virtex2P" and FAMILY/="virtex2") generate
         comp: entity work.sub_inst(struct)
               generic map(
                     DW => DW,
                     NUM => ISUB
               )
               port map(
                     clk    => clk,
                     en     => pix_en,
                     op_a   => sub_ops_a(ISUB),
                     op_b   => sub_ops_b(ISUB),
                     res    => sub_res(ISUB)
               );
      end generate;
   end generate;
end generate;
--end instantiation

-- registering comparison results
GEN_CMPR_OLD_L0:
for IT in 0 to TAP-NI-1 generate
   GEN_CMPR_OLD_L1:
   for INS in 0 to NI-1 generate
   begin
      process(clk)
      begin
      if (clk'event and clk='1') then      
         if (pix_en='1') then
            cmpr_out_old(IT)(INS) <= sub_res(INS*(TAP-NI)+IT)(DW);
         end if;
      end if;
      end process;
   end generate;
end generate;


GEN_CMPR_NEW_EQ:
for I in 0 to NI-1 generate
   process(clk)
   begin
   if (clk'event and clk='1') then      
      if (pix_en='1') then
         cmpr_out_new(I)(I) <= '1';
      end if;
   end if;
   end process;
end generate;

GEN_CMPR_NEW:
for IN0 in 0 to NI-2 generate
   L1: for IN1 in IN0+1 to NI-1 generate
      process(clk)
      begin
      if (clk'event and clk='1') then      
         if (pix_en='1') then
            cmpr_out_new(IN1)(IN0) <= sub_res( ((TAP-NI)*NI)+ (NI*IN0-(IN0*IN0+IN0)/2) + IN1-IN0-1)(DW);
         end if;
      end if;
      end process;
   end generate;
end generate;

----------------------------------------------------------------
-- generate shift register for the new sample(s)
--    using the old-new comparisons
--    and the new-new comparisons
GEN_C_SHREG_N0:
for IT in 0 to NI-1 generate
   GEN_C_SHREG_N1:
   for IB in 0 to TAP-1 generate
      CASE0: 
      if (not(IB < NI)) generate
         process(clk)
         begin
         if (clk'event and clk='1') then
            if (pix_en='1') then
               c_shreg(IT)(IB) <= not cmpr_out_old(IB-NI)(IT);
            end if;
         end if;
         end process;
      end generate;
      CASE1:
      if ((IB < NI) AND not(IB<=IT)) generate
         process(clk)
         begin
         if (clk'event and clk='1') then
            if (pix_en='1') then
               c_shreg(IT)(IB) <= not cmpr_out_new(IB)(IT);
            end if;
         end if;
         end process;
      end generate;
      CASE2:
      if (IB <= IT) generate
         process(clk)
         begin
         if (clk'event and clk='1') then
            if (pix_en='1') then
               c_shreg(IT)(IB) <= cmpr_out_new(IT)(IB);
            end if;
         end if;
         end process;
      end generate;
   end generate;
end generate;

-- generate shift register for the old samples
--    using the old-new comparisons
GEN_C_SHREG_O:
for I in NI to TAP-1 generate
begin
   process(clk)
   begin
   if (clk'event and clk='1') then
      if (pix_en='1') then
         c_shreg(I) <= c_shreg(I-NI)(TAP-NI-1 downto 0) & cmpr_out_old(I-NI);
      end if;
   end if;
   end process;
end generate;


-- 1CNT - counting the 1s in the c_shreg registers
-- or in other words, summing the bits
-- for the new samples adder tree is used
-- samples are masked according to their validity
GEN_1CNT_NEW:
for IT in 0 to NI-1 generate
   masked_full(IT)  <= c_shreg(IT) AND mask_cell;
   count_new: entity work.cntr1_new(rtl)
   generic map(
      TAP       => TAP,
      TAPb      => TAPb,
      RST_VAL   => 0,
      ADDER_INS => ADDER_INS
   )
   port map(
      clk         => clk,
      en          => pix_en,
      data        => masked_full(IT),
      res         => one_cntr(IT)
   );
end generate;

-- 1CNT - counting the 1s in the c_shreg registers
-- or in other words, summing the bits
-- for the old samples increment/decrement modules are used
GEN_1CNTR_OLD:
for IT in 0 to TAP-NI-1 generate
   MI: for IB in 0 to NI-1 generate   
      masked_in(IT)(IB)  <= cmpr_out_old(IT)(IB) AND mask_ni(IB);
   end generate;
   masked_out(IT)  <= c_shreg(IT)(TAP-1 downto TAP-NI) AND mask_ni;
   count_old: entity work.cntr1_old(rtl)
   generic map(
      TAP       => TAP,
      NI        => NI,
      TAPb      => TAPb,
      RST_VAL   => 0,
      ADDER_INS => ADDER_INS
   )
   port map(
      clk         => clk,
      rst         => rst,
      en          => pix_en,
      data_old    => masked_out(IT),
      data_new    => masked_in(IT),
      cntr_in     => one_cntr(IT),
      res         => one_cntr(IT+NI)
   );
end generate;


-------------------------------------------------------------------------------------------------------
-- comparing the output of the 1CNT registers with the input rank
process(clk)
variable CELL_VALID : std_logic_vector(TAP-1 downto 0);
begin
for I in 0 to TAP-1 loop
   if ((I mod WH_virt) < (WH_virt-WH_real)) then
      CELL_VALID(I) := '0';
   else
      CELL_VALID(I) := '1';
   end if;
end loop;
if (clk'event and clk='1') then
   if (pix_en='1') then
      for IT in 0 to TAP-1 loop
         if (CELL_VALID(IT)='1') then
            if (one_cntr(IT) = rank) then
               rank_cmpr(IT) <= '1';
            else
               rank_cmpr(IT) <= '0';
            end if;
         else
            rank_cmpr(IT) <= '0';
         end if;
      end loop;
   end if;
end if;
end process;


-- N-to-1 decoder generates the output address from the rank_cmpr result
-- two stages
process(clk)
variable cell_addr : array_NI_by_INTEGER;
variable cell_num:   array_rank0_t1;
variable INS : integer;
variable CELL_VALID : std_logic_vector(TAP-1 downto 0);
begin
for I in 0 to TAP-1 loop
   if ((I mod WH_virt) < (WH_virt-WH_real)) then
      CELL_VALID(I) := '0';
   else
      CELL_VALID(I) := '1';
   end if;
end loop;
if (clk'event and clk='1') then
   if (pix_en='1') then
      for INS in 0 to NI-1 loop
         cell_addr(INS) := 0;
         for IR in 0 to RR_PER_NI-1 loop
            cell_num(INS)(IR) := CONST_0(TAPb-1 downto 0);
         end loop;
      end loop;
      for IT in 0 to TAP-1 loop
         INS := IT mod NI;
         if (CELL_VALID(IT)='1') then
            if (rank_cmpr(IT)='1') then
               cell_num(INS)(cell_addr(INS)/RANK_PER_STAGE) := CONV_STD_LOGIC_VECTOR(cell_addr(INS), TAPb);
            end if;
         end if;
         cell_addr(INS) := cell_addr(INS) + 1;
      end loop;
      for INS in 0 to NI-1 loop
         for IR in 0 to RR_PER_NI-1 loop
            ranked_cell0(INS)(IR) <= cell_num(INS)(IR);
         end loop;
      end loop;
   end if;
end if;
end process;

process(clk)
variable CELL_VALID : std_logic_vector(TAP-1 downto 0);
variable sel : array_mux_sel;
begin
for I in 0 to TAP-1 loop
   if ((I mod WH_virt) < (WH_virt-WH_real)) then
      CELL_VALID(I) := '0';
   else
      CELL_VALID(I) := '1';
   end if;
end loop;
if (clk'event and clk='1') then
   if (pix_en='1') then
      for IR in 0 to MUX0_REGS-1 loop
         sel(IR) := CONST_0(LOG2_CEIL(NI)-1 downto 0);      
      end loop;
      for IT in 0 to TAP-1 loop
         if (CELL_VALID(IT)='1') then
            if (rank_cmpr(IT)='1') then 
               sel(IT/MUX_OR_INS) := CONV_STD_LOGIC_VECTOR((IT mod NI), LOG2_CEIL(NI));        
            end if;
         end if;
      end loop;
      for IR in 0 to MUX0_REGS-1 loop
         mux_sel0(IR) <= sel(IR);
      end loop;
   end if;
end if;
end process;

process(clk)
variable sel : std_logic_vector(LOG2_CEIL(NI)-1 downto 0);
begin
if (clk'event and clk='1') then
   if (pix_en='1') then
      sel := CONST_0(LOG2_CEIL(NI)-1 downto 0);
      for IR in 0 to MUX0_REGS-1 loop
         sel := sel or mux_sel0(IR);
      end loop;
      mux_sel1 <= sel;
   end if;
end if;
end process;
out_mux <= mux_sel1;


process(clk)
variable rank1_or : array_rank1;
begin
if (clk'event and clk='1') then
   if (pix_en='1') then
      for INS in 0 to NI-1 loop
         rank1_or(INS) := CONST_0(TAPb-1 downto 0);
      end loop;
      for INS in 0 to NI-1 loop
         for IR in 0 to RR_PER_NI-1 loop
            rank1_or(INS) := rank1_or(INS) OR ranked_cell0(INS)(IR);
         end loop;
      end loop;
      for INS in 0 to NI-1 loop
         ranked_cell1(INS) <= rank1_or(INS);
      end loop;
   end if;
end if;
end process;

GEN_OUT_ADDR:
for INS in 0 to NI-1 generate
   out_addr((INS+1)*TAPb-1 downto INS*TAPb) <= ranked_cell1(INS);
end generate;

-------------------------------------------------------------------------------------------------------

end rtl;
