--*******************************************************************
-- Copyright(C) 2005 by Xilinx, Inc. All rights reserved.
-- This text/file contains proprietary, confidential
-- information of Xilinx, Inc., is distributed under license
-- from Xilinx, Inc., and may be used, copied and/or
-- disclosed only pursuant to the terms of a valid license
-- agreement with Xilinx, Inc. Xilinx hereby grants you
-- a license to use this text/file solely for design, simulation,
-- implementation and creation of design files limited
-- to Xilinx devices or technologies. Use with non-Xilinx
-- devices or technologies is expressly prohibited and
-- immediately terminates your license unless covered by
-- a separate agreement.
--
-- Xilinx is providing this design, code, or information
-- "as is" solely for use in developing programs and
-- solutions for Xilinx devices. By providing this design,
-- code, or information as one possible implementation of
-- this feature, application or standard, Xilinx is making no
-- representation that this implementation is free from any
-- claims of infringement. You are responsible for
-- obtaining any rights you may require for your implementation.
-- Xilinx expressly disclaims any warranty whatsoever with
-- respect to the adequacy of the implementation, including
-- but not limited to any warranties or representations that this
-- implementation is free from claims of infringement, implied
-- warranties of merchantability or fitness for a particular
-- purpose.
--
-- Xilinx products are not intended for use in life support
-- appliances, devices, or systems. Use in such applications are
-- expressly prohibited.
--
-- This copyright and support notice must be retained as part
-- of this text at all times. (c) Copyright 2005 Xilinx, Inc.
-- All rights reserved.
--
-- Title - Xil_FVG.vhd
-- Author(s) - GZ, Xilinx
-- Creation - 04/03/2006
--
-- $RCSfile: fvg_y.vhd,v $ $Revision: 1.3 $ $Date: 2006/06/30 19:01:23 $
--
-- ************************************************************************
--
-- Description: Filter Value Generator (FVG) for the 2D-rank project
--
-- Generalized conversion:
-- Y  =  RCOEF * R + G + BCOEF * B
--
-- ******************************************************************  
library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_arith.all;
use ieee.std_logic_unsigned.all;

LIBRARY work;
USE work.genxlib_utils.ALL;

entity Xil_FVG is   
  generic ( 
    FAMILY_HAS_MAC: integer:= 0;
    FABRIC_ADDS   : integer:= 1;      -- Adders are implemented using logic fabric based adders
    IWIDTH        : integer:= 8;      -- MAX 17 to fit into DSP48 or mult18x18s
    CWIDTH        : integer:= 16;     -- MAX 17 to fit into DSP48 or mult18x18s 
    MWIDTH        : integer:= 16;     -- For non-DSP48 based implementations; limits adder carry chain length
    OWIDTH        : integer:= 10;
    RCOEF         : integer:= 33417;   -- 0.29931/0.587 = 0.50989; round( 0.50989 * pow2(CWIDTH) ) = 33417
    BCOEF         : integer:= 12728);  -- 0.11376/0.587 = 0.19421; round( 0.19421 * pow2(CWIDTH) ) = 12728 
  port (                               
    R             : in std_logic_vector(IWIDTH-1 downto 0);   
    G             : in std_logic_vector(IWIDTH-1 downto 0);
    B             : in std_logic_vector(IWIDTH-1 downto 0);
    Y             : out std_logic_vector(OWIDTH-1 downto 0);  -- Y  = a(R-G) + G + b(B-G)
    clk           : in std_logic;
    ce            : in std_logic := '1';
    sclr          : in std_logic := '0');
end Xil_FVG;

architecture rtl of Xil_FVG is            

  constant MODULE_LATENCY : integer := 5;  -- ADDER_DELAY is set to 1, MULT_DELAY is 2
  constant RCOEFvec       : std_logic_vector(CWIDTH downto 0) := conv_std_logic_vector(RCOEF,CWIDTH+1); 
  constant BCOEFvec       : std_logic_vector(CWIDTH downto 0) := conv_std_logic_vector(BCOEF,CWIDTH+1);
  constant PADD_INSERT    : integer := min(CWIDTH-1, max(0 , CWIDTH+IWIDTH-OWIDTH));
-- Low level constants 
  constant logic0         : std_logic := '0';

-- signal declarations
-- Only for the DSP48 based implementation
  signal PADD_ROUND       : std_logic_vector(CWIDTH-1 downto 0);
  signal g_delay          : std_logic_vector(IWIDTH-1 downto 0);
  signal b_delay          : std_logic_vector(IWIDTH downto 0);
  signal r_padded         : std_logic_vector(IWIDTH downto 0);
  signal rm               : std_logic_vector(IWIDTH+CWIDTH+1 downto 0);
  signal g_padded         : std_logic_vector(IWIDTH+CWIDTH+1  downto 0);
  signal bm               : std_logic_vector(IWIDTH+CWIDTH+1 downto 0);
  signal rm_plus_g        : std_logic_vector(IWIDTH+CWIDTH+1 downto 0); 
  signal y_int            : std_logic_vector(IWIDTH+CWIDTH+1 downto 0);  

begin        

---------------------------------------------------------------------
--  Calculate the R*RCOEF, B*BCOEF
--------------------------------------------------------------------- 

  -- padd G component with zeros and the rounding constant 0..011..11 
  -- in order to align decimal points of G and rm, and at the same time 
  -- add ~ 0.5 to the final output reult (thereby implementing rounding
  -- rather then truncation when dropping bits from MWIDTH+2 to OWIDTH
  padd_rnd_0: if (PADD_INSERT<CWIDTH) generate PADD_ROUND(CWIDTH-1 downto PADD_INSERT) <= (others =>'0'); end generate;
  padd_rnd_1: if (PADD_INSERT>0) generate PADD_ROUND(PADD_INSERT-1 downto 0) <= (others =>'1'); end generate;
  
  del_G : entity work.delay(rtl) -- Delay matching: G is delayed so it is in sync with rm 
  generic map ( width => IWIDTH, delay => 2 )   
  port map ( clk => clk, ce => ce, d => G, q => g_delay);

  del_B : entity work.delay(rtl) -- Delay matching: B is delayed so its product bn is in sync with rm_plus_G 
  generic map ( width => IWIDTH, delay => 1 )   
  port map ( clk => clk, ce => ce, d => B, q => b_delay(IWIDTH-1 downto 0) );
  b_delay(IWIDTH) <= '0'; -- turn B to unsigned positive
  
  r_padded(IWIDTH-1 downto 0) <= R;
  r_padded(IWIDTH) <= '0'; -- turn R to unsigned positive

  g_padded(CWIDTH-1 downto 0) <= PADD_ROUND;  
  g_padded(IWIDTH+CWIDTH-1 downto CWIDTH) <= g_delay;
  g_padded(IWIDTH+CWIDTH)   <= '0'; -- turn G to unsigned positive
  g_padded(IWIDTH+CWIDTH+1) <= '0'; 
  
  sp3_v2_v2p_mult: if (FAMILY_HAS_MAC=0) generate
     mult_r: entity work.mult(rtl)
      generic map ( 
        iwidtha => IWIDTH+1, 
        iwidthb => CWIDTH+1,
        delay   => 2)   
      port map (
        clk     => clk,
        ce      => ce,
        sclr    => sclr,
        a       => R_padded,
        b       => RCOEFvec,
        p       => rm);
    
     mult_b: entity work.mult(rtl)
      generic map ( 
        iwidtha => IWIDTH+1, 
        iwidthb => CWIDTH+1,
        delay   => 2)   
      port map (
        clk     => clk,
        ce      => ce,
        sclr    => sclr,
        a       => B_delay,
        b       => BCOEFvec,
        p       => bm);
      
    add_G_and_rm : entity work.radd_sub_sclr(rtl) -- rm_plus_g = R*RCOEF + g
      generic map ( 
        width => MWIDTH, 
        add   => true,
        fabric=> 1,
        delay => 1)   
      port map (
        a     => rm(IWIDTH+CWIDTH downto IWIDTH+CWIDTH-MWIDTH+1),
        b     => g_padded(IWIDTH+CWIDTH downto IWIDTH+CWIDTH-MWIDTH+1),
        s     => rm_plus_g(MWIDTH downto 0),
        c_in  => logic0,
        clk   => clk,
        ce    => ce,
        sclr  => sclr);  

    rm_plus_g_and_bm : entity work.radd_sub_sclr(rtl) -- rm_plus_g = R*RCOEF + g
      generic map ( 
        width => MWIDTH, 
        add   => true,
        fabric=> 1,
        delay => 1)   
      port map (
        a     => bm(IWIDTH+CWIDTH downto IWIDTH+CWIDTH-MWIDTH+1),
        b     => rm_plus_g(MWIDTH-1 downto 0),
        s     => y_int(MWIDTH downto 0),
        c_in  => logic0,
        clk   => clk,
        ce    => ce,
        sclr  => sclr);  

    Y <= y_int(MWIDTH-1 downto MWIDTH-OWIDTH);
        
  end generate;
  
  v4_mac1: if (FAMILY_HAS_MAC=1) generate         -- DSP48 based implementation
    mac_r: entity work.mac(rtl)                   -- rm_plus_g = RCOEFF * R + G +  rounding const
      generic map (                               
        IWIDTHA   => IWIDTH+1, 
        IWIDTHB   => CWIDTH+1,   
        OWIDTH    => IWIDTH+CWIDTH+2,
        ROUND_MODE=> 0,
        HAS_C     => 1)
      port map (
        clk       => clk,
        ce        => ce,
        sclr      => sclr,
        a         => R_padded,
        b         => RCOEFvec,
        c         => g_padded,
        p         => rm_plus_g);
        
    mac_b: entity work.mac(rtl)                -- Y_int = BCOEFF * B + G +  rm_plus_g
      generic map (                               
        IWIDTHA   => IWIDTH+1, 
        IWIDTHB   => CWIDTH+1,   
        OWIDTH    => IWIDTH+CWIDTH+2,
        ROUND_MODE=> 0,
        HAS_C     => 1)
      port map (
        clk       => clk,
        ce        => ce,
        sclr      => sclr,
        a         => B_delay,
        b         => BCOEFvec,
        c         => rm_plus_g,
        p         => y_int);
        
    Y <= y_int(IWIDTH+CWIDTH downto IWIDTH+CWIDTH+1-OWIDTH);

  end generate;
  
end rtl; 

