--------------------------------------------------------------------------------
-- Copyright(C) 2005 by Xilinx, Inc. All rights reserved.
-- This text/file contains proprietary, confidential
-- information of Xilinx, Inc., is distributed under license
-- from Xilinx, Inc., and may be used, copied and/or
-- disclosed only pursuant to the terms of a valid license
-- agreement with Xilinx, Inc. Xilinx hereby grants you
-- a license to use this text/file solely for design, simulation,
-- implementation and creation of design files limited
-- to Xilinx devices or technologies. Use with non-Xilinx
-- devices or technologies is expressly prohibited and
-- immediately terminates your license unless covered by
-- a separate agreement.
--
-- Xilinx is providing this design, code, or information
-- "as is" solely for use in developing programs and
-- solutions for Xilinx devices. By providing this design,
-- code, or information as one possible implementation of
-- this feature, application or standard, Xilinx is making no
-- representation that this implementation is free from any
-- claims of infringement. You are responsible for
-- obtaining any rights you may require for your implementation.
-- Xilinx expressly disclaims any warranty whatsoever with
-- respect to the adequacy of the implementation, including
-- but not limited to any warranties or representations that this
-- implementation is free from claims of infringement, implied
-- warranties of merchantability or fitness for a particular
-- purpose.
--
-- Xilinx products are not intended for use in life support
-- appliances, devices, or systems. Use in such applications are
-- expressly prohibited.
--
-- This copyright and support notice must be retained as part
-- of this text at all times. (c) Copyright 2005 Xilinx, Inc.
-- All rights reserved.
--------------------------------------------------------------------------------
--
-- Filename - rank2d_top.vhd
-- Author - P. Szanto
-- Creation - 27 June 2006
--
-- Description -
--
-- Top-level entity of the 2D rank filter.
-- This module
--    generates the required control signals
--       for the operation (addresses, enable, valid signals)
--       the output control signals (pix_en_out, hs_out, vs_out)
--    instantiates sub-modules
--       line buffer
--       FVG
--       delay line
--       filter core
--
--
-- $RCSfile: rank2d_top.vhd,v $ $Revision: 1.3 $ $Date: 2006/06/30 19:01:24 $
-- 
--------------------------------------------------------------------------------
library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.STD_LOGIC_ARITH.ALL;
use IEEE.STD_LOGIC_UNSIGNED.ALL;

use work.rank2d_utils.ALL;
use work.rank2d_latency.ALL;

entity rank2d_top is
generic(DATA_WIDTH_CH0     : integer := 8;         -- Bit width of color channel 0
        DATA_WIDTH_CH1     : integer := 8;         -- Bit width of color channel 1
        DATA_WIDTH_CH2     : integer := 8;         -- Bit width of color channel 2
        DATA_WIDTH_FILTER  : integer := 10;        -- Filter data bit width
        WINDOW_WIDTH       : integer := 5;         -- Horizontal size of filter window
        WINDOW_HEIGHT      : integer := 5;         -- Vertical size of filter window
        NEW_INPUTS         : integer := 1;         -- new input samples/clk for the filter core
        MAX_HORIZONTAL_RES : integer := 2048;      -- maximum number of pixels per line
        Y_GENERATOR_TYPE   : integer := 0;         -- Filter Value Generator Type; 0: sum of CC0, CC1 and CC2; 1: compute luminance; 2: use CC0
        FAMILY             : string  := "Virtex4"  -- FPGA family
        );
port   (clk       : in  std_logic;
        rst       : in  std_logic;
        ce        : in  std_logic;
        pix_en_in : in  std_logic;
        rank      : in  std_logic_vector(GET_TAPb (WINDOW_HEIGHT*WINDOW_WIDTH,WINDOW_HEIGHT,NEW_INPUTS)-1 downto 0);
        cc0_in    : in  std_logic_vector(DATA_WIDTH_CH0-1 downto 0);
        cc1_in    : in  std_logic_vector(DATA_WIDTH_CH1-1 downto 0);
        cc2_in    : in  std_logic_vector(DATA_WIDTH_CH2-1 downto 0);
        vs_in     : in  std_logic;
        hs_in     : in  std_logic;
        cc0_out   : out std_logic_vector(DATA_WIDTH_CH0-1 downto 0);
        cc1_out   : out std_logic_vector(DATA_WIDTH_CH1-1 downto 0);
        cc2_out   : out std_logic_vector(DATA_WIDTH_CH2-1 downto 0);
        pix_en_out: out std_logic;
        vs_out    : out std_logic;
        hs_out    : out std_logic
);
end rank2d_top;

architecture rtl of rank2d_top is

constant WW             : integer := WINDOW_WIDTH;

constant WH_real        : integer := WINDOW_HEIGHT;
constant WH_virt        : integer := DIV_CEIL(WH_real, NEW_INPUTS)*NEW_INPUTS;       -- virtual filter window vertical size
constant TAPr           : integer := WH_real*WW;                     -- real tap number
constant TAP            : integer := WH_virt*WW;                     -- virtual TAP number
constant TAPb           : integer := LOG2_CEIL(TAP+1);               -- data width to represent TAP
constant CYCLES_PER_COL : integer := WH_virt/NEW_INPUTS;             -- input cycles to process a new filter column


-- FPGA FAMILY DEPENDANT CONSTANTS
--------------------------------------------------------------------
constant ADDER_INS      : integer := GET_ADDER_INS(FAMILY);
constant SRL_DEPTH      : integer := GET_SRL_DEPTH(FAMILY);
constant SRL_ADDRb      : integer := LOG2_CEIL(SRL_DEPTH);
constant EN_TO_BRAM_OUT : integer := GET_BRAM_DELAY(FAMILY);                                 -- delay from enable input to LBUFF output reg
-- Enables usage of instantiated primitives
constant USE_INST_DL    : integer := 1;      -- use instantiated SRLs in delay line
constant USE_INST_COMP  : integer := 0;      -- use instantiated, RLOCed comparators in filter core
-- XST requires ISE_INST_DL to be set to "1", otherwise it cannot infer SRLs
-- Comparator instantiation is set to "0" by default, as par with "-t 8" and "-t 9" options typically do a great job placing them
-- If you have timing problems, try setting this constant to "1"
--------------------------------------------------------------------

constant CONST_0  : std_logic_vector(127 downto 0) := x"00000000000000000000000000000000";
constant CONST_1  : std_logic_vector(127 downto 0) := x"FFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF";

constant DWF            : integer := DATA_WIDTH_CH0 + DATA_WIDTH_CH1 + DATA_WIDTH_CH2;

-----------------
constant CNT1_WIDTH     : integer := GET_1CNT_WIDTH (0, TAP, 0, CYCLES_PER_COL);
constant CNT1_LATENCY   : integer := GET_1CNT_STAGES(CNT1_WIDTH, ADDER_INS); 
constant RANK_LATENCY   : integer := 4 + CNT1_LATENCY + 3;                       -- filter core latency
constant FVG_LATENCY    : integer := GET_FVG_LATENCY(Y_GENERATOR_TYPE);          -- Filter Value Generator latency
------------------
constant FIFO_DEPTH     : integer := DIV_CEIL(TAP, NEW_INPUTS) + RANK_LATENCY + FVG_LATENCY;     -- Delay line depth
constant FIFO_NUMSRL    : integer := DIV_CEIL(FIFO_DEPTH, SRL_DEPTH);            -- number of SRLs to form a single delay line
constant FIFO_ADDRb     : integer := LOG2_CEIL(FIFO_NUMSRL)+SRL_ADDRb;           -- address width for the delay line
constant FIFO_ADDR_DLb  : integer := MAX(1, FIFO_ADDRb-SRL_ADDRb);
constant LBUFF_DEPTH    : integer := MAX_HORIZONTAL_RES;                         -- line buffer depth
constant LBUFF_ADDR_W   : integer := LOG2_CEIL(MAX_HORIZONTAL_RES);              -- address width for the line buffer


constant FILTER_EN_DL   : integer := 3 + 1 + FVG_LATENCY + CYCLES_PER_COL;       -- (blockRAM out reg + mux + FVG)   +  CYCLES_PER_COL for disabling

constant HSYNC_DL       : integer := 1 + CYCLES_PER_COL + FVG_LATENCY + RANK_LATENCY - 1;

constant COLOR_REGS     : integer := 1 + DIV_CEIL(2, DIV_CEIL(WH_real, NEW_INPUTS));  -- number of color (input delay) registers to compensate memory latency

constant ADDR_MOD       : integer := RANK_LATENCY + FVG_LATENCY - 1;             -- address modifier for the delay line address (base address for the current data)

type array_LBUFF_OUT    is array (WH_real-2 downto 0) of STD_LOGIC_VECTOR(DWF-1 downto 0);
type array_WVv_by_DWF   is array (WH_virt-1 downto 0) of STD_LOGIC_VECTOR(DWF-1 downto 0);
type array_NI_by_DWF    is array (NEW_INPUTS-1 downto 0) of STD_LOGIC_VECTOR(DWF-1 downto 0);
type array_NI_by_DW     is array (NEW_INPUTS-1 downto 0) of STD_LOGIC_VECTOR(DATA_WIDTH_FILTER-1 downto 0);
type array_wr_addr      is array (EN_TO_BRAM_OUT-1 downto 0) of STD_LOGIC_VECTOR(LBUFF_ADDR_W-1 downto 0);
type array_wr_sel       is array (EN_TO_BRAM_OUT-1 downto 0) of STD_LOGIC_VECTOR(WH_real-2 downto 0);
type array_COLOR_NEW    is array (EN_TO_BRAM_OUT-1 downto 0) of STD_LOGIC_VECTOR(DWF-1 downto 0);
type array_DLOUTB       is array (NEW_INPUTS-1 downto 0) of std_logic_vector(FIFO_NUMSRL*DWF-1 downto 0);
type array_DL_ADDR      is array (NEW_INPUTS-1 downto 0) of STD_LOGIC_VECTOR(TAPb-1 downto 0);
type array_DL_ADDR_MOD  is array (NEW_INPUTS-1 downto 0) of STD_LOGIC_VECTOR(FIFO_ADDRb-1 downto 0);
type array_SRL_by_DWF   is array (FIFO_NUMSRL-1 downto 0) of STD_LOGIC_VECTOR(DWF-1 downto 0);
type array_dl_out16     is array (NEW_INPUTS-1 downto 0) of array_SRL_by_DWF;

signal pix_en_r         : std_logic := '0';
signal pix_en           : std_logic := '0';
signal pix_en_dl_cpc    : std_logic_vector(CYCLES_PER_COL downto 0) := (others => '0');
signal pix_en_cpc_r     : std_logic := '0';
signal pix_en_cpc       : std_logic := '0';

signal color_in         : std_logic_vector(DWF-1 downto 0);
signal color_out        : std_logic_vector(DWF-1 downto 0);

-- control signals
signal hs_in_dl         : std_logic_vector(1 downto 0) := "00";
signal hs_fall          : std_logic;
signal vs_in_dl         : std_logic_vector(1 downto 0) := "00";

-- signals for the line buffers
signal lbuff_out_w      : array_LBUFF_OUT;
signal lbuff_out_reg    : array_LBUFF_OUT;
signal lbuff_addr_rd_mux : STD_LOGIC_VECTOR(LBUFF_ADDR_W-1 downto 0);
signal lbuff_addr_rd_din : STD_LOGIC_VECTOR(LBUFF_ADDR_W-1 downto 0);
signal lbuff_addr_rd_en  : STD_LOGIC;
signal lbuff_addr_rd    : STD_LOGIC_VECTOR(LBUFF_ADDR_W-1 downto 0) := CONST_0(LBUFF_ADDR_W-1 downto 0);
signal lbuff_addr_wr    : array_wr_addr;
signal lbuff_addr_wr_w  : STD_LOGIC_VECTOR(LBUFF_ADDR_W-1 downto 0);
signal lbuff_wr_sel     : STD_LOGIC_VECTOR(WH_real-2 downto 0) := CONST_1(0) & CONST_0(WH_real-3 downto 0);
signal lbuff_wr_en      : STD_LOGIC_VECTOR(WH_real-2 downto 0);
signal lbuff_din        : std_logic_vector(DWF-1 downto 0);
signal color_new        : array_COLOR_NEW;
signal lbuff_out_bus    : array_WVv_by_DWF;
signal color_mux_sel    : STD_LOGIC_VECTOR(LOG2_CEIL(CYCLES_PER_COL)-1 downto 0);
signal color_muxed      : array_NI_by_DWF;
signal filter_val       : array_NI_by_DW; 

-- signals for the delay lines
signal delay_line_din : array_NI_by_DWF;
signal delay_line_out_B : array_DLOUTB;

-- filter core
signal filter_din   : STD_LOGIC_VECTOR(DATA_WIDTH_FILTER*NEW_INPUTS-1 downto 0);
signal filter_addr  : STD_LOGIC_VECTOR(NEW_INPUTS*TAPb-1 downto 0);
signal filter_mux   : STD_LOGIC_VECTOR(LOG2_CEIL(NEW_INPUTS)-1 downto 0);
signal filter_mux_r : STD_LOGIC_VECTOR(LOG2_CEIL(NEW_INPUTS)-1 downto 0);
signal dl_addr : array_DL_ADDR;
signal dl_addr_mod : array_DL_ADDR_MOD;
signal dl_addr_mod_dl : STD_LOGIC_VECTOR(FIFO_ADDR_DLb-1 downto 0);
signal delay_line_out     : array_dl_out16;
signal delay_line_out_reg : array_dl_out16;
signal filter_valid : STD_LOGIC;
signal filter_valid_dl : STD_LOGIC;
signal filter_valid_dl_din : STD_LOGIC;
signal out_valid : STD_LOGIC;
signal out_valid_din : STD_LOGIC;
signal en_dl0 : STD_LOGIC_VECTOR(EN_TO_BRAM_OUT-1 downto 0);
signal en_dl1 : STD_LOGIC_VECTOR(HSYNC_DL-1 downto 0) := (others => '0');
--signal en_dl2 : STD_LOGIC_VECTOR(1 downto 0);
signal rst_dl0 : STD_LOGIC_VECTOR(EN_TO_BRAM_OUT-1 downto 0);
signal rst_dl1 : STD_LOGIC_VECTOR(HSYNC_DL-1 downto 0) := (others => '0');
signal rst_out : STD_LOGIC;
signal hs_dl0 : STD_LOGIC_VECTOR(EN_TO_BRAM_OUT-1 downto 0);
signal hs_dl1 : STD_LOGIC_VECTOR(HSYNC_DL-1 downto 0);
signal hs_dl2 : STD_LOGIC_VECTOR(1 downto 0);
signal vs_dl0 : STD_LOGIC_VECTOR(EN_TO_BRAM_OUT-1 downto 0);
signal vs_dl1 : STD_LOGIC_VECTOR(HSYNC_DL-1 downto 0);
signal vs_dl2 : STD_LOGIC_VECTOR(1 downto 0);
signal t_pix_en_dl : std_logic_vector(1 downto 0);
signal t_cyc_cpc : std_logic_vector(CYCLES_PER_COL downto 0);
signal t_filter_valid_dl : std_logic_vector(1 + FVG_LATENCY + RANK_LATENCY + CYCLES_PER_COL downto 0);

-- attribute types
attribute syn_keep : boolean;
attribute syn_direct_enable : boolean;
attribute syn_maxfan : integer;
attribute syn_srlstyle : string;
attribute syn_preserve : boolean;
attribute syn_useenables : boolean;
attribute syn_replicate : boolean;

attribute syn_keep of lbuff_addr_rd_din : signal is true;
attribute syn_direct_enable of lbuff_addr_rd_en : signal is true;

attribute syn_maxfan of pix_en_cpc_r, pix_en_r : signal is 100;
attribute syn_srlstyle of color_new, lbuff_din : signal is "registers";

attribute syn_srlstyle of t_filter_valid_dl, rst_dl1, hs_dl1, vs_dl1 : signal is "select_srl";

signal t_pix_en_cpc_r : std_logic;
signal t_pix_en_cpc   : std_logic;
attribute syn_preserve of t_pix_en_cpc_r : signal is true;
attribute syn_keep of t_pix_en_cpc : signal is true;

-- In V2/V2P/SP3, BlockRAMs are relatively far away from each other.
-- To reduce routing delay, BRAM input signals are replicated
attribute syn_maxfan of lbuff_addr_rd, lbuff_addr_wr_w, lbuff_wr_sel, lbuff_din : signal is GET_BRAM_IN_FAN(FAMILY, TAP);
--attribute syn_maxfan of lbuff_addr_rd_din : signal is 1;

begin

color_in <= cc0_in & cc1_in & cc2_in;

process(clk)
begin
if (clk'event and clk='1') then
   if (rst='1') then
      pix_en_r <= '0';
   elsif (pix_en_in='1' and ce='1') then
      pix_en_r <= '1';
   else
      pix_en_r <= '0';
   end if;
end if;
end process;
pix_en <= pix_en_r AND ce;

process(clk)
begin
if (clk'event and clk='1') then
   if (ce='1') then
      pix_en_dl_cpc(CYCLES_PER_COL downto 1) <= pix_en_dl_cpc(CYCLES_PER_COL-1 downto 0);
      if (pix_en_in='1') then
         pix_en_dl_cpc(0) <= '1';
      else
         pix_en_dl_cpc(0) <= '0';
      end if;
   end if;
end if;
end process;


process(clk)
begin
if (clk'event and clk='1') then
  if (rst='1') then   
      pix_en_cpc_r <= '0';
   elsif (ce='1') then
      if (pix_en_in='1') then
         pix_en_cpc_r <= '1';
      elsif (pix_en_dl_cpc(CYCLES_PER_COL-1)='1') then
         pix_en_cpc_r <= '0';
      end if;
   end if;
end if;
end process;
pix_en_cpc <= pix_en_cpc_r AND ce;

-- detecting falling edge of hsync and vsync
-- the second pixel after HS is handled as the first pixel of the new line
process(clk)
begin
if (clk'event and clk='1') then
   if (pix_en_in='1' and ce='1') then
      hs_in_dl <= hs_in_dl(0) & hs_in;
   end if;
end if;
end process;

process(clk)
begin
if (clk'event and clk='1') then
   if (pix_en_in='1' and ce='1') then
      hs_fall <= hs_in_dl(0) and not hs_in;
   end if;
end if;
end process;

process(clk)
begin
if (clk'event and clk='1') then
   if (pix_en_in='1' and ce='1' and hs_in = '1') then
      vs_in_dl <= vs_in_dl(0) & vs_in;
   end if;
end if;
end process;


-- line buffer address generation, line buffers
-- selecting the line buffer to write to
-- input cycle shift register
--lbuff_addr_rd_mux <= CONST_0(LBUFF_ADDR_W-1 downto 0) when (hs_in_dl="10") else lbuff_addr_rd;
lbuff_addr_rd_mux <= CONST_1(LBUFF_ADDR_W-1 downto 0) when (hs_fall='1') else lbuff_addr_rd;
lbuff_addr_rd_din <= lbuff_addr_rd_mux + 1;
lbuff_addr_rd_en  <= rst or pix_en;

process(clk)
begin
if (clk'event and clk='1') then
   if (rst='1') then
      lbuff_addr_rd <= CONST_0(LBUFF_ADDR_W-1 downto 0);
   elsif (lbuff_addr_rd_en='1') then
--      if (hs_in_dl = "10") then
--         lbuff_addr_rd <= CONST_0(LBUFF_ADDR_W-1 downto 0);
--      else
--         lbuff_addr_rd <= lbuff_addr_rd + 1;
--      end if;
      lbuff_addr_rd <= lbuff_addr_rd_din;
   end if;
end if;
end process;


-- write address for the line buffer(s)
-- it is the appropriately delayed read address

-- lbuff_wr_sel(0) is synchronized with the read address, then 
-- delayed as the write address
process(clk)
begin
if (clk'event and clk='1') then
   if (rst='1') then
      for IR in 0 to EN_TO_BRAM_OUT-2 loop
         lbuff_addr_wr(IR)  <= CONST_0(LBUFF_ADDR_W-1 downto 0);
      end loop;
   elsif (pix_en='1') then
      for IR in 0 to EN_TO_BRAM_OUT-2 loop
      if (IR=0) then
         lbuff_addr_wr(IR)  <= lbuff_addr_rd;
      else   
         lbuff_addr_wr(IR)  <= lbuff_addr_wr(IR-1);
      end if;
      end loop;
   end if;
end if;
end process;

process(clk)
begin
if (clk'event and clk='1') then
   if (rst='1') then
      lbuff_wr_sel <= CONST_1(0) & CONST_0(WH_real-3 downto 0);
   elsif (pix_en='1' AND hs_in_dl = "10") then
      if (vs_in_dl="10") then
         lbuff_wr_sel <= CONST_1(0) & CONST_0(WH_real-3 downto 0);
      else
         lbuff_wr_sel <= lbuff_wr_sel(0) & lbuff_wr_sel(WH_real-2 downto 1);
      end if;
   end if;
end if;
end process;

-- delayed color input
process(clk)
begin
if (clk'event and clk='1') then
   for IR in 0 to EN_TO_BRAM_OUT-1 loop
      if (IR=0) then      
         if (pix_en_in = '1' and ce='1') then
            color_new(0) <= color_in;
         end if;
      else
         if (pix_en='1') then
            color_new(IR) <= color_new(IR-1);
         end if;
      end if;
   end loop;
end if;
end process;

-- line buffers
-- WW-1 MAX_HORIZONTAL_RES deep memories
--lbuff_addr_wr_w <= lbuff_addr_wr(EN_TO_BRAM_OUT-2);

process(clk)
begin
if (clk'event and clk='1') then
   if (pix_en='1') then
      lbuff_addr_wr_w <= lbuff_addr_wr(EN_TO_BRAM_OUT-3);
      lbuff_din       <= color_new(EN_TO_BRAM_OUT-2);
   end if;
end if;
end process;

GEN_LBUFFS:
for IB in 0 to WH_real-2 generate
   lbuff_wr_en(IB) <= '1' when ( lbuff_wr_sel(IB)='1') else '0';

   lbuff: entity work.lbuff_mem(rtl)
   generic map
          (DWF       => DWF,
           LW_MAX    => MAX_HORIZONTAL_RES,
           FAMILY    => FAMILY
   )
   port map
          (clk       => clk,
           rst       => rst,
           en        => pix_en,
           wr_en     => lbuff_wr_en(IB),
           addr_rd   => lbuff_addr_rd,
           addr_wr   => lbuff_addr_wr_w,
           din       => lbuff_din,
           dout      => lbuff_out_w(IB)
   );
   process(clk)
   begin
   if (clk'event and clk='1') then
      if (pix_en='1') then
         lbuff_out_reg(IB) <= lbuff_out_w(IB);
      end if;
   end if;
   end process;
end generate;


-- joined data "bus"
-- WV-1 data from the line buffers
-- then the new data
-- then padding data (if WH_virt > WH_real). Padding data is generated in a way that minimizes the number of multiplexers required.
GEN_LBUFF_BUS:
for IWV in 0 to WH_virt-1 generate
   NEW_SAMPLE: 
   if (IWV = WH_real-1) generate
      lbuff_out_bus(IWV) <= color_new(EN_TO_BRAM_OUT-1);
   end generate;
   PADDING:
   if (IWV > WH_real-1) generate
      lbuff_out_bus(IWV) <= lbuff_out_reg((IWV mod WH_real) + WH_real mod NEW_INPUTS);
   end generate;
   LBUFF_SAMPLES:
   if (IWV < WH_real-1) generate
      lbuff_out_bus(IWV) <= lbuff_out_reg(IWV);
   end generate;
end generate;

-- multiplexer to select NEW_INPUTS data from lbuff_out_bus
process(clk)
begin
   if (clk'event and clk='1') then
      if (pix_en='1') then
         color_mux_sel <= CONST_0(LOG2_CEIL(CYCLES_PER_COL)-1 downto 0);
      elsif (pix_en_cpc='1') then
         color_mux_sel <= color_mux_sel + 1;
      end if;
   end if;
end process;

-- multiplexed data
-- feeds delay line and FVG
GEN_RGB_MUX:
for INS in 0 to NEW_INPUTS-1 generate
process (clk)
begin
if (clk'event and clk='1') then
   if (pix_en_cpc='1') then
      color_muxed(INS) <= lbuff_out_bus(CONV_INTEGER(color_mux_sel)*NEW_INPUTS + INS);
   end if;
end if;
end process;
end generate;


-- FVG: generating values for filtering
-- generating filter value, NEW_INPUTS FVG units
GEN_FVGs:
for INS in 0 to NEW_INPUTS-1 generate
   FVG_IS_SUM: if (Y_GENERATOR_TYPE=0) generate
      FVG: entity work.fvg_sum(rtl)
      generic map(
              DW_CC0     => DATA_WIDTH_CH0,
              DW_CC1     => DATA_WIDTH_CH1,
              DW_CC2     => DATA_WIDTH_CH2,
              DW         => DATA_WIDTH_FILTER
            )
      port map  (
              clk        => clk,
              en         => pix_en_cpc,
              cc0_in     => color_muxed(INS)(DATA_WIDTH_CH0+DATA_WIDTH_CH1+DATA_WIDTH_CH2-1 downto DATA_WIDTH_CH1+DATA_WIDTH_CH2),
              cc1_in     => color_muxed(INS)(DATA_WIDTH_CH1+DATA_WIDTH_CH2-1 downto DATA_WIDTH_CH2),
              cc2_in     => color_muxed(INS)(DATA_WIDTH_CH2-1 downto 0),
              filter_val => filter_val(INS)
            );
      end generate;
   FVG_IS_Y: if (Y_GENERATOR_TYPE=1) generate
      FVG: entity work.Xil_FVG(rtl)
        generic map( 
          FAMILY_HAS_MAC => GET_FAMILY_HAS_MAC(FAMILY),
          FABRIC_ADDS    => GET_FABRIC_ADDS(FAMILY),
          IWIDTH         => 8,
          CWIDTH         => 16,
          MWIDTH         => 16,
          OWIDTH         => DATA_WIDTH_FILTER,
          RCOEF          => 33417,
          BCOEF          => 12728)
        port map(                               
          R              => color_muxed(INS)(DATA_WIDTH_CH0+DATA_WIDTH_CH1+DATA_WIDTH_CH2-1 downto DATA_WIDTH_CH1+DATA_WIDTH_CH2),
          G              => color_muxed(INS)(DATA_WIDTH_CH1+DATA_WIDTH_CH2-1 downto DATA_WIDTH_CH2),
          B              => color_muxed(INS)(DATA_WIDTH_CH2-1 downto 0),
          Y              => filter_val(INS),
          clk            => clk,
          ce             => pix_en_cpc,
          sclr           => '0');
   end generate;
   FVG_IS_BYP: if (Y_GENERATOR_TYPE=2) generate
      FVG: entity work.fvg_yuv(rtl)
      generic map(
              DW_CC0     => DATA_WIDTH_CH0,
              DW_CC1     => DATA_WIDTH_CH1,
              DW_CC2     => DATA_WIDTH_CH2,
              DW         => DATA_WIDTH_FILTER
            )
      port map  (
              clk        => clk,
              cc0_in     => color_muxed(INS)(DATA_WIDTH_CH0+DATA_WIDTH_CH1+DATA_WIDTH_CH2-1 downto DATA_WIDTH_CH1+DATA_WIDTH_CH2),
              cc1_in     => color_muxed(INS)(DATA_WIDTH_CH1+DATA_WIDTH_CH2-1 downto DATA_WIDTH_CH2),
              cc2_in     => color_muxed(INS)(DATA_WIDTH_CH2-1 downto 0),
              filter_val => filter_val(INS)
            );
   end generate;
end generate;


-- delay lines, implemented with SRLs; number of SRLs: NEW_INPUTS*FIFO_DEPTH/SRL_DEPTH
-- each SRL feds a register, so SRL output routing/logic delay is not increased with logic
GEN_DELAY_LINE:
for INS in 0 to NEW_INPUTS-1 generate
   delay_line_din(INS) <= color_muxed(NEW_INPUTS-1-INS);
   GEN_HDL:
   if (USE_INST_DL=0) generate
      dl: entity work.delay_line(rtl)
      generic map
             (SRL_DEPTH => SRL_DEPTH,
              SRL_ADDRb => SRL_ADDRb,
              NUM_SRL   => FIFO_NUMSRL,
              WIDTH     => DWF
            )
      port map   
             (clk       => clk,
              we        => pix_en_cpc,
              addr      => dl_addr_mod(INS)(SRL_ADDRb-1 downto 0),
              din       => delay_line_din(INS),
              dout      => delay_line_out_B(INS)
            );
   end generate;
   GEN_PRIM:
   if (USE_INST_DL=1) generate
      GEN_V5:
      if (FAMILY="virtex5") generate
         dl: entity work.delay_line_srl32(struct)
         generic map
                (SRL_DEPTH => SRL_DEPTH,
                 SRL_ADDRb => SRL_ADDRb,
                 NUM_SRL   => FIFO_NUMSRL,
                 WIDTH     => DWF
               )
         port map   
                (clk       => clk,
                 we        => pix_en_cpc,
                 addr      => dl_addr_mod(INS)(SRL_ADDRb-1 downto 0),
                 din       => delay_line_din(INS),
                 dout      => delay_line_out_B(INS)
               );
      end generate;
      GEN_NON_V5:
      if (FAMILY/="virtex5") generate
         dl: entity work.delay_line_srl16(struct)
         generic map
                (SRL_DEPTH => SRL_DEPTH,
                 SRL_ADDRb => SRL_ADDRb,
                 NUM_SRL   => FIFO_NUMSRL,
                 WIDTH     => DWF
               )
         port map   
                (clk       => clk,
                 we        => pix_en_cpc,
                 addr      => dl_addr_mod(INS)(SRL_ADDRb-1 downto 0),
                 din       => delay_line_din(INS),
                 dout      => delay_line_out_B(INS)
               );
      end generate;
   end generate;
   GEN_DL_OUT:
   for ISRL in 0 to FIFO_NUMSRL-1 generate
      delay_line_out(INS)(ISRL) <= delay_line_out_B(INS)((ISRL+1)*DWF-1 downto ISRL*DWF);   
   end generate;   
end generate;


-- instantiating filter core
GEN_FILTER_DIN:
for INS in 0 to NEW_INPUTS-1 generate
   filter_din(DATA_WIDTH_FILTER*(INS+1)-1 downto DATA_WIDTH_FILTER*INS) <= filter_val(NEW_INPUTS-1-INS);
end generate;

filter: entity work.filter_core(rtl)
   generic map
          (DW             => DATA_WIDTH_FILTER,
           TAPr           => TAPr,
           TAPb           => TAPb,
           WH_real        => WH_real,
           NI             => NEW_INPUTS,
           ADDER_INS      => ADDER_INS,
           FAMILY         => FAMILY,
           USE_INST_COMP  => USE_INST_COMP)
   port map
          (clk       => clk,
           rst       => rst,
           ce        => '1',
           pix_en    => pix_en_cpc,
           m_ni_en   => en_dl1(2),
           m_cell_en => en_dl1(3),
           rank      => rank,
           nd_port   => filter_din,
           out_addr  => filter_addr,
           out_mux   => filter_mux
         );

-- generating delay line address (filter core output + modification value)  
GEN_FIFO_ADDR:
for INS in 0 to NEW_INPUTS-1 generate
   dl_addr(INS)        <= filter_addr(TAPb*(INS+1)-1 downto TAPb*INS);
   DL_NO_ADDR_EXT: if (TAPb=FIFO_ADDRb) generate
      dl_addr_mod(INS) <= dl_addr(INS) + ADDR_MOD;      
   end generate;
   DL_ADDR_EXT: if (TAPb<FIFO_ADDRb) generate
      dl_addr_mod(INS) <= (CONST_0(FIFO_ADDRb-TAPb-1 downto 0) & dl_addr(INS)) + ADDR_MOD;      
   end generate;
   DL_ADDR_REM: if (TAPb>FIFO_ADDRb) generate
      dl_addr_mod(INS) <= dl_addr(INS)(FIFO_ADDRb-1 downto 0) + ADDR_MOD;      
   end generate;
end generate;

-- delay line SRL output registers
GEN_DL_OUT_REGS0: 
for INS in 0 to NEW_INPUTS-1 generate
   GEN_DL_OUT_REGS1: 
   for ISRL in 0 to FIFO_NUMSRL-1 generate
      process(clk)
      begin
      if (clk'event and clk='1') then
         if (filter_valid='1' and ce='1') then
            delay_line_out_reg(INS)(ISRL) <= delay_line_out(INS)(ISRL);
         end if;
      end if;
      end process;
   end generate;
end generate;
process(clk)
begin   
   if (clk'event and clk='1') then
      if (filter_valid='1' and ce='1') then
         filter_mux_r <= filter_mux;
      end if;
   end if;
end process;
SINGLE_SRL: if (FIFO_NUMSRL=1) generate
process(clk)
begin
   if (clk'event and clk='1') then
      if (filter_valid='1' and ce='1') then
         dl_addr_mod_dl <= CONST_0(FIFO_ADDRb-SRL_ADDRb-1 downto 0);
      end if;
   end if;
end process;
end generate;

MULT_SRLs: if (FIFO_NUMSRL/=1) generate
process(clk)
begin
   if (clk'event and clk='1') then
      if (filter_valid='1' and ce='1') then
         dl_addr_mod_dl <= dl_addr_mod(CONV_INTEGER(filter_mux))(FIFO_ADDRb-1 downto SRL_ADDRb);
      end if;
   end if;
end process;
end generate;


-- multiplexing output registers
process(clk)
begin
if (clk'event and clk='1') then
   if (ce='1') then
      color_out <= delay_line_out_reg(CONV_INTEGER(filter_mux_r))( CONV_INTEGER(dl_addr_mod_dl) );
   end if;
end if;
end process;

cc0_out <= color_out(DATA_WIDTH_CH0+DATA_WIDTH_CH1+DATA_WIDTH_CH2-1 downto DATA_WIDTH_CH1+DATA_WIDTH_CH2);
cc1_out <= color_out(DATA_WIDTH_CH1+DATA_WIDTH_CH2-1 downto DATA_WIDTH_CH2);
cc2_out <= color_out(DATA_WIDTH_CH2-1 downto 0);

-- output valid signals
process(clk)
begin
if (clk'event and clk='1') then
   if (rst='1') then
      t_pix_en_dl <= "00";
   elsif (pix_en='1') then
      t_pix_en_dl <= t_pix_en_dl(0) & '1';
   end if;
end if;
end process;

--process(clk)
--begin
--if (clk'event and clk='1') then
--   if (pix_en_cpc='1') then
--      if ((t_pix_en_dl(0)='1' and pix_en='1') or t_cyc_cpc(CYCLES_PER_COL-1)='1') then   
--         t_cyc_cpc(0) <= '1';
--      else
--         t_cyc_cpc(0) <= '0';
--      end if;
--      t_cyc_cpc(CYCLES_PER_COL downto 1) <= t_cyc_cpc(CYCLES_PER_COL-1 downto 0);
--   end if;
--end if;
--end process;

process(clk)
begin
if (clk'event and clk='1') then
   if (rst='1') then
      t_cyc_cpc <= CONST_0(CYCLES_PER_COL downto 0);
   elsif (pix_en_cpc='1') then
      if ((t_pix_en_dl(0)='1' and pix_en='1')) then   
         t_cyc_cpc(0) <= '1';
      else
         t_cyc_cpc(0) <= '0';
      end if;
      t_cyc_cpc(CYCLES_PER_COL downto 1) <= t_cyc_cpc(CYCLES_PER_COL-1 downto 0);
   end if;
end if;
end process;

process(clk)
begin
if (clk'event and clk='1') then
   if (rst='1') then   
      t_pix_en_cpc_r <= '0';
   elsif (ce='1') then
      if (pix_en_in='1') then
         t_pix_en_cpc_r <= '1';
      elsif (pix_en_dl_cpc(CYCLES_PER_COL-1)='1') then
         t_pix_en_cpc_r <= '0';
      end if;
   end if;
end if;
end process;
t_pix_en_cpc <= t_pix_en_cpc_r AND ce AND t_pix_en_dl(1);

process(clk)
begin
if (clk'event and clk='1') then
--   if (rst='1') then
--      t_filter_valid_dl <= CONST_0(1 + FVG_LATENCY + RANK_LATENCY + CYCLES_PER_COL downto 0);      
   if (t_pix_en_cpc='1') then
      t_filter_valid_dl <= t_filter_valid_dl(1 + FVG_LATENCY + RANK_LATENCY + CYCLES_PER_COL-1 downto 0) & t_cyc_cpc(CYCLES_PER_COL-1);      
   end if;
end if;
end process;

GEN_VALID_0:
if (WH_virt=NEW_INPUTS) generate
   filter_valid <= '1' when (t_filter_valid_dl(1 + FVG_LATENCY + RANK_LATENCY - 1)='1' AND pix_en='1') else '0';
end generate;
GEN_VALID_1:
if (WH_virt/=NEW_INPUTS) generate
   filter_valid <= t_filter_valid_dl(1 + FVG_LATENCY + RANK_LATENCY - 1) AND pix_en_cpc;
end generate;


filter_valid_dl_din <= '0' when (rst='1' or rst_out='1') else filter_valid;
out_valid_din       <= '0' when (rst='1' or rst_out='1') else filter_valid_dl;
process(clk)
begin
if (clk'event and clk='1') then
   if (ce='1') then
      filter_valid_dl <= filter_valid_dl_din;
      out_valid       <= out_valid_din;
   end if;
end if;
end process;
pix_en_out <= out_valid;

--- GENERATING HSYNC output
process(clk)
begin
if (clk'event and clk='1') then
   for IR in 0 to EN_TO_BRAM_OUT-1 loop
      if (IR=0) then      
         if (pix_en_in = '1' and ce='1') then
            en_dl0(0)  <= pix_en_in;
            hs_dl0(0)  <= hs_in;
            vs_dl0(0)  <= vs_in;
         end if;
      else
         if (pix_en='1') then
            en_dl0(IR)  <= en_dl0(IR-1);
            hs_dl0(IR)  <= hs_dl0(IR-1);
            vs_dl0(IR)  <= vs_dl0(IR-1);
         end if;
      end if;
   end loop;
end if;
end process;

process(clk)
begin
if (clk'event and clk='1') then
   for IR in 0 to EN_TO_BRAM_OUT-1 loop
      if (IR=0) then      
         if (rst='1') then
            rst_dl0(IR) <= '1';
       elsif (pix_en = '1' and rst='0') then
            rst_dl0(IR) <= '0';
         end if;
      else
         if (pix_en='1') then
            rst_dl0(IR) <= rst_dl0(IR-1);
         end if;
      end if;
   end loop;
end if;
end process;

process(clk)
begin
if (clk'event and clk='1') then
   if (pix_en_cpc='1') then
      en_dl1  <= en_dl1(HSYNC_DL-2 downto 0) & en_dl0(EN_TO_BRAM_OUT-1);
      rst_dl1 <= rst_dl1(HSYNC_DL-2 downto 0) & rst_dl0(EN_TO_BRAM_OUT-1);
      hs_dl1  <= hs_dl1(HSYNC_DL-2 downto 0) & hs_dl0(EN_TO_BRAM_OUT-1);
      vs_dl1  <= vs_dl1(HSYNC_DL-2 downto 0) & vs_dl0(EN_TO_BRAM_OUT-1);
   end if;
end if;
end process;

process(clk)
begin
if (clk'event and clk='1') then
   if (rst_out='1') then
      hs_dl2(0)  <= '0';
      vs_dl2(0)  <= '0';
      hs_dl2(1) <= '0';
      vs_dl2(1) <= '0';
   elsif (ce='1') then   
      if (filter_valid='1') then
         hs_dl2(0)  <= hs_dl1(HSYNC_DL-1);
         vs_dl2(0)  <= vs_dl1(HSYNC_DL-1);
      end if;
      hs_dl2(1)  <= hs_dl2(0);
      vs_dl2(1)  <= vs_dl2(0);
   end if;
end if;
end process;
hs_out <= hs_dl2(1);
vs_out <= vs_dl2(1);

process(clk)
begin
if (clk'event and clk='1') then
   if (rst='1') then
      rst_out <= '1';
   elsif (rst_dl1(HSYNC_DL-2)='1') then
      rst_out <= '0';
   end if;
end if;
end process;


end rtl;
